@@ -12,11 +12,9 @@
 
 """
 
-from cssselect.parser import SelectorError
-from cssselect.xpath import Translator
+from cssselect.parser import SelectorError, SelectorSyntaxError
+from cssselect.xpath import GenericTranslator, HTMLTranslator, ExpressionError
 
 
-VERSION = '0.2'
+VERSION = '0.3'
 __version__ = VERSION
-
-css_to_xpath = Translator().css_to_xpath
