@@ -22,8 +22,7 @@ import unittest
 
 from lxml import html
 from cssselect.parser import tokenize, parse, parse_series, SelectorSyntaxError
-from cssselect.xpath import Translator, ExpressionError
-from cssselect import css_to_xpath
+from cssselect.xpath import GenericTranslator, HTMLTranslator, ExpressionError
 
 
 class TestCssselect(unittest.TestCase):
@@ -32,7 +31,9 @@ class TestCssselect(unittest.TestCase):
                   for item in tokenize('E > f[a~="y\\"x"]')]
         assert tokens == [
             "Symbol('E', 0)",
+            "Token(' ', 1)",
             "Token('>', 2)",
+            "Token(' ', 3)",
             "Symbol('f', 4)",
             "Token('[', 5)",
             "Symbol('a', 6)",
@@ -47,6 +48,13 @@ class TestCssselect(unittest.TestCase):
                 assert repr(parse(other)) == result
             return result.replace("(u'", "('")
 
+        assert parse_many('*') == 'Element[*]'
+        assert parse_many('*|*') == 'Element[*]'
+        assert parse_many('*|foo') == 'Element[foo]'
+        assert parse_many('foo|*') == 'Element[foo|*]'
+        assert parse_many('foo|bar') == 'Element[foo|bar]'
+        # This will never match, but it is valid:
+        assert parse_many('#foo#bar') == 'Hash[Hash[Element[*]#foo]#bar]'
         assert parse_many(
             'div>.foo',
             'div> .foo',
@@ -74,26 +82,30 @@ class TestCssselect(unittest.TestCase):
         assert parse_many('td ::first') == (
             'CombinedSelector[Element[td] '
                 '<followed> Pseudo[Element[*]::first]]')
-        assert parse_many('a[name]') == (
+        assert parse_many('a[name]', 'a[ name\t]') == (
             'Attrib[Element[a][name]]')
         assert parse_many('a [name]') == (
             'CombinedSelector[Element[a] <followed> Attrib[Element[*][name]]]')
         assert parse_many('a[rel="include"]') == (
             "Attrib[Element[a][rel = String('include', 6)]]")
+        assert parse_many('a[rel = include]') == (
+            "Attrib[Element[a][rel = Symbol('include', 8)]]")
         assert parse_many("a[hreflang |= 'en']") == (
             "Attrib[Element[a][hreflang |= String('en', 14)]]")
         assert parse_many('div:nth-child(10)') == (
-            'Function[Element[div]:nth-child(10)]')
+            "Function[Element[div]:nth-child(Symbol('10', 14))]")
+        assert parse_many(':nth-child(2n+2)') == (
+            "Function[Element[*]:nth-child(Symbol('2n+2', 11))]")
         assert parse_many('div:nth-of-type(10)') == (
-            'Function[Element[div]:nth-of-type(10)]')
+            "Function[Element[div]:nth-of-type(Symbol('10', 16))]")
         assert parse_many('div div:nth-of-type(10) .aclass') == (
-            'CombinedSelector[CombinedSelector[Element[div] '
-                '<followed> Function[Element[div]:nth-of-type(10)]] '
+            'CombinedSelector[CombinedSelector[Element[div] <followed> '
+                "Function[Element[div]:nth-of-type(Symbol('10', 20))]] "
                 '<followed> Class[Element[*].aclass]]')
         assert parse_many('label:only') == (
             'Pseudo[Element[label]:only]')
         assert parse_many('a:lang(fr)') == (
-            'Function[Element[a]:lang(Element[fr])]')
+            "Function[Element[a]:lang(Symbol('fr', 7))]")
         assert parse_many('div:contains("foo")') == (
             "Function[Element[div]:contains(String('foo', 13))]")
         assert parse_many('div#foobar') == (
@@ -120,25 +132,61 @@ class TestCssselect(unittest.TestCase):
         assert get_error('html/body/a') == (
             "Unexpected symbol: '/' at [Symbol('html', 0)] -> None")
         assert get_error(' ') == (
-            "Expected selector, got 'None' at [] -> None")
+            "Expected selector, got 'None' at [Token(' ', 0)] -> None")
         assert get_error('div, ') == (
             "Expected selector, got 'None' at "
-            "[Symbol('div', 0), Token(',', 3)] -> None")
+            "[Symbol('div', 0), Token(',', 3), Token(' ', 4)] -> None")
         assert get_error(' , div') == (
             "Expected selector, got ',' at "
-            "[] -> Token(',', 1)")
+            "[Token(' ', 0)] -> Token(',', 1)")
         assert get_error('p, , div') == (
             "Expected selector, got ',' at "
-            "[Symbol('p', 0), Token(',', 1)] -> Token(',', 3)")
+            "[Symbol('p', 0), Token(',', 1), Token(' ', 2)] -> Token(',', 3)")
         assert get_error('div > ') == (
             "Expected selector, got 'None' at "
-            "[Symbol('div', 0), Token('>', 4)] -> None")
-        assert get_error(' > div') == (
-            "Expected selector, got '>' at [] -> Token('>', 1)")
+            "[Symbol('div', 0), Token(' ', 3), Token('>', 4), Token(' ', 5)]"
+            " -> None")
+        assert get_error('  > div') == (
+            "Expected selector, got '>' at [Token(' ', 0)] -> Token('>', 2)")
+        assert get_error('foo|#bar') == (
+            "Expected symbol or '*', got '#' at "
+            "[Symbol('foo', 0), Token('|', 3), "
+            "Token('#', 4)] -> Symbol('bar', 5)")
+        assert get_error('#.foo') == (
+            "Expected symbol, got '.' at "
+            "[Token('#', 0), Token('.', 1)] -> Symbol('foo', 2)")
+        assert get_error('.#foo') == (
+            "Expected symbol, got '#' at "
+            "[Token('.', 0), Token('#', 1)] -> Symbol('foo', 2)")
+        assert get_error(':#foo') == (
+            "Expected symbol, got '#' at "
+            "[Token(':', 0), Token('#', 1)] -> Symbol('foo', 2)")
+        assert get_error('[*]') == (
+            "Expected '|', got ']' at "
+            "[Token('[', 0), Token('*', 1)] -> Token(']', 2)")
+        assert get_error('[foo|]') == (
+            "Expected symbol, got ']' at "
+            "[Token('[', 0), Symbol('foo', 1), Token('|', 4), Token(']', 5)]"
+            " -> None")
+        assert get_error('[#]') == (
+            "Expected symbol or '*', got '#' at "
+            "[Token('[', 0), Token('#', 1)] -> Token(']', 2)")
+        assert get_error('[foo=#]') == (
+            "Expected string or symbol, got '#' at "
+            "[Token('[', 0), Symbol('foo', 1), Token('=', 4), Token('#', 5)]"
+            " -> Token(']', 6)")
+        assert get_error(':nth-child()') == (
+            "Expected argument, got ')' at "
+            "[Token(':', 0), Symbol('nth-child', 1), Token('(', 10)]"
+            " -> Token(')', 11)")
+        assert get_error('[href]a') == (
+            "Expected selector, got 'a' at "
+            "[Token('[', 0), Symbol('href', 1), Token(']', 5)]"
+            " -> Symbol('a', 6)")
 
     def test_translation(self):
         def xpath(css):
-            return str(Translator().xpath(parse(css)))
+            return str(GenericTranslator().xpath(parse(css)))
 
         assert xpath('*') == "*"
         assert xpath('E') == "e"
@@ -189,6 +237,10 @@ class TestCssselect(unittest.TestCase):
             "e[not(*) and not(normalize-space())]")
         assert xpath('E:root') == (
             "e[not(parent::*)]")
+        assert xpath('E:contains("foo")') == (
+            "e[contains(string(.), 'foo')]")
+        assert xpath('E:contains(foo)') == (
+            "e[contains(string(.), 'foo')]")
         assert xpath('E.warning') == (
             "e[@class and contains("
                "concat(' ', normalize-space(@class), ' '), ' warning ')]")
@@ -214,7 +266,7 @@ class TestCssselect(unittest.TestCase):
         else:
             css = '.a\xc1b'.decode('ISO-8859-1')
 
-        xpath = css_to_xpath(css)
+        xpath = GenericTranslator().css_to_xpath(css)
         assert css[1:] in xpath
         xpath = xpath.encode('ascii', 'xmlcharrefreplace').decode('ASCII')
         assert xpath == (
@@ -222,6 +274,7 @@ class TestCssselect(unittest.TestCase):
             "concat(' ', normalize-space(@class), ' '), ' a&#193;b ')]")
 
     def test_quoting(self):
+        css_to_xpath = GenericTranslator().css_to_xpath
         assert css_to_xpath('*[aval="\'"]') == (
             '''descendant-or-self::*[@aval = "'"]''')
         assert css_to_xpath('*[aval="\'\'\'"]') == (
@@ -233,6 +286,7 @@ class TestCssselect(unittest.TestCase):
 
     def test_unicode_escapes(self):
         # \22 == '"'  \20 == ' '
+        css_to_xpath = GenericTranslator().css_to_xpath
         assert css_to_xpath(r'*[aval="\'\22\'"]') == (
             '''descendant-or-self::*[@aval = concat("'",'"',"'")]''')
         assert css_to_xpath(r'*[aval="\'\22 2\'"]') == (
@@ -256,17 +310,24 @@ class TestCssselect(unittest.TestCase):
         sort_key = dict(
             (el, count) for count, el in enumerate(document.getiterator())
         ).__getitem__
+        css_to_xpath = GenericTranslator().css_to_xpath
+        html_css_to_xpath = HTMLTranslator().css_to_xpath
 
-        def select_ids(selector):
+        def select_ids(selector, html_only):
             xpath = css_to_xpath(selector)
             items = document.xpath(xpath)
+            if html_only:
+                assert items == []
+                xpath = html_css_to_xpath(selector)
+                items = document.xpath(xpath)
             items.sort(key=sort_key)
             return [element.get('id', 'nil') for element in items]
 
-        def pcss(main, *selectors):
-            result = select_ids(main)
+        def pcss(main, *selectors, **kwargs):
+            html_only = kwargs.pop('html_only', False)
+            result = select_ids(main, html_only)
             for selector in selectors:
-                assert select_ids(selector) == result
+                assert select_ids(selector, html_only) == result
             return result
 
         all_ids = pcss('*')
@@ -296,9 +357,7 @@ class TestCssselect(unittest.TestCase):
         # FIXME: I'm not 100% sure this is right:
         assert pcss('li:nth-child(3n+1)') == [
             'first-li', 'fourth-li', 'seventh-li']
-        # FIXME: I'm not sure if nth-last-child(1) or nth-last-child(1)
-        # should be equivalent to nth-last-child()
-        assert pcss('li:nth-last-child()', 'li:nth-last-child(0)') == [
+        assert pcss('li:nth-last-child(0)') == [
             'seventh-li']
         assert pcss('li:nth-last-child(2n)', 'li:nth-last-child(even)') == [
             'second-li', 'fourth-li', 'sixth-li']
@@ -310,14 +369,23 @@ class TestCssselect(unittest.TestCase):
         assert pcss('ol:nth-last-of-type(1)') == ['first-ol']
         assert pcss('span:only-child') == ['foobar-span']
         assert pcss('li div:only-child') == ['li-div']
-        assert pcss('div *:only-child') == ['li-div', 'foobar-span']
+        assert pcss('div *:only-child') == [
+            'li-div', 'checkbox-disabled', 'foobar-span']
         self.assertRaises(ExpressionError, pcss, 'p *:only-of-type')
+        self.assertRaises(ExpressionError, pcss, 'p:lang(fr)')
         assert pcss('p:only-of-type') == ['paragraph']
         assert pcss('a:empty') == ['name-anchor']
         assert pcss('li:empty') == [
             'third-li', 'fourth-li', 'fifth-li', 'sixth-li', 'seventh-li']
         assert pcss(':root', 'html:root') == ['html']
         assert pcss('li:root', '* :root') == []
+        assert pcss('*:contains("link")') == [
+            'html', 'nil', 'outer-div', 'tag-anchor', 'nofollow-anchor']
+        assert pcss('*:contains("LInk")') == []  # case sensitive
+        assert pcss('*:contains("e")') == [
+            'html', 'nil', 'outer-div', 'first-ol', 'first-li',
+            'paragraph', 'p-em']
+        assert pcss('*:contains("E")') == []  # case-sensitive
         assert pcss('.a', '.b', '*.a', 'ol.a') == ['first-ol']
         assert pcss('.c', '*.c') == ['first-ol', 'third-li', 'fourth-li']
         assert pcss('ol *.c', 'ol li.c', 'li ~ li.c', 'ol > li.c') == [
@@ -334,11 +402,19 @@ class TestCssselect(unittest.TestCase):
         assert pcss('ol#first-ol *:last-child') == ['li-div', 'seventh-li']
         assert pcss('#outer-div:first-child') == ['outer-div']
         assert pcss('#outer-div :first-child') == [
-            'name-anchor', 'first-li', 'li-div', 'p-b']
+            'name-anchor', 'first-li', 'li-div', 'p-b', 'checkbox-disabled']
+        assert pcss('a[href]') == ['tag-anchor', 'nofollow-anchor']
+        assert pcss(':link', html_only=True) == pcss('a[href]')
+        assert pcss(':checked', html_only=True) == ['checkbox-checked']
+        assert pcss(':disabled', html_only=True) == [
+            'fieldset', 'checkbox-disabled']
+        assert pcss(':enabled', html_only=True) == [
+            'checkbox-unchecked', 'checkbox-checked']
 
     def test_select_shakespeare(self):
         document = html.document_fromstring(HTML_SHAKESPEARE)
         body = document.xpath('//body')[0]
+        css_to_xpath = GenericTranslator().css_to_xpath
 
         try:
             basestring_ = basestring
@@ -362,6 +438,7 @@ class TestCssselect(unittest.TestCase):
         ## searching the body.
         #assert count('*') == 252
         assert count('*') == 246
+        assert count('div:contains(CELIA)') == 26
         assert count('div:only-child') == 22 # ?
         assert count('div:nth-child(even)') == 106
         assert count('div:nth-child(2n)') == 106
@@ -429,7 +506,13 @@ c"></li>
  </ol>
  <p id="paragraph">
    <b id="p-b">hi</b> <em id="p-em">there</em>
-   <b id="p-b2">guy</b></p>
+   <b id="p-b2">guy</b>
+   <input type="checkbox" id="checkbox-unchecked">
+   <input type="checkbox" id="checkbox-checked" checked="checked">
+   <fieldset id="fieldset" disabled="disabled">
+     <input type="checkbox" id="checkbox-disabled">
+   </fieldset>
+ </p>
  <ol id="second-ol">
  </ol>
 </div>
