@@ -78,7 +78,8 @@ class XPathExpr(object):
         if self.element == '*':
             # We weren't doing a test anyway
             return
-        self.add_condition("name() = %s" % xpath_literal(self.element))
+        self.add_condition(
+            "name() = %s" % GenericTranslator.xpath_literal(self.element))
         self.element = '*'
 
     def add_star_prefix(self):
@@ -126,27 +127,13 @@ class XPathExprOr(XPathExpr):
 
 split_at_single_quotes = re.compile("('+)").split
 
-def xpath_literal(s):
-    if isinstance(s, Element):
-        # This is probably a symbol that looks like an expression...
-        s = s._format_element()
-    else:
-        s = _unicode(s)
-    if "'" not in s:
-        s = "'%s'" % s
-    elif '"' not in s:
-        s = '"%s"' % s
-    else:
-        s = "concat(%s)" % ','.join([
-            (("'" in part) and '"%s"' or "'%s'") % part
-            for part in split_at_single_quotes(s) if part
-            ])
-    return s
-
 
 #### Translation
 
-class Translator(object):
+class GenericTranslator(object):
+    """
+    Translator for "generic" XML documents.
+    """
     combinator_mapping = {
         ' ': 'descendant',
         '>': 'child',
@@ -173,12 +160,12 @@ class Translator(object):
 
         .. sourcecode:: pycon
 
-            >>> from cssselect import css_to_xpath
-            >>> exrpession = css_to_xpath('div.content')
-            >>> exrpession
-            "descendant-or-self::div[contains(concat(' ', normalize-space(@class), ' '), ' content ')]"
+            >>> from cssselect import GenericTranslator
+            >>> expression = GenericTranslator().css_to_xpath('div.content')
+            >>> print(expression)
+            descendant-or-self::div[@class and contains(concat(' ', normalize-space(@class), ' '), ' content ')]
 
-        The resulting expression can be used with lxml’s `XPath engine`_:
+        The resulting expression can be used with lxml's `XPath engine`_:
 
         .. _XPath engine: http://lxml.de/xpathxslt.html#xpath
 
@@ -201,6 +188,24 @@ class Translator(object):
         xpath.add_prefix(prefix or '')
         return _unicode(xpath)
 
+    @staticmethod
+    def xpath_literal(s):
+        if isinstance(s, Element):
+            # This is probably a symbol that looks like an expression...
+            s = s._format_element()
+        else:
+            s = _unicode(s)
+        if "'" not in s:
+            s = "'%s'" % s
+        elif '"' not in s:
+            s = '"%s"' % s
+        else:
+            s = "concat(%s)" % ','.join([
+                (("'" in part) and '"%s"' or "'%s'") % part
+                for part in split_at_single_quotes(s) if part
+                ])
+        return s
+
     def xpath(self, parsed_selector, prefix=None):
         """Translate any parsed selector object."""
         type_name = type(parsed_selector).__name__
@@ -232,7 +237,7 @@ class Translator(object):
         method = getattr(self, method, None)
         if not method:
             raise ExpressionError(
-                "The pseudo-class :%r is unknown" % function.name)
+                "The pseudo-class :%s() is unknown" % function.name)
         return method(self.xpath(function.selector), function)
 
     def xpath_pseudo(self, pseudo):
@@ -240,8 +245,9 @@ class Translator(object):
         method = 'xpath_%s_pseudo' % pseudo.ident.replace('-', '_')
         method = getattr(self, method, None)
         if not method:
+            # TODO: better error message for pseudo-elements?
             raise ExpressionError(
-                "The pseudo-class :%r is unknown" % pseudo.ident)
+                "The pseudo-class :%s is unknown" % pseudo.ident)
         return method(self.xpath(pseudo.selector))
 
 
@@ -257,23 +263,19 @@ class Translator(object):
             name = '@' + selector.attrib
         else:
             name = '@%s:%s' % (selector.namespace, selector.attrib)
-        return method(self.xpath(selector.selector),
-                      name, operator, selector.value)
+        return method(self.xpath(selector.selector), name, selector.value)
 
     def xpath_class(self, class_selector):
         """Translate a class selector."""
-        # FIXME: is this HTML-specific?
+        # .foo is defined as [class~=foo] in the spec.
         xpath = self.xpath(class_selector.selector)
-        xpath.add_condition(
-            "@class and "
-            "contains(concat(' ', normalize-space(@class), ' '), %s)"
-            % xpath_literal(' ' + class_selector.class_name + ' '))
-        return xpath
+        return self.xpath_attrib_includes(
+            xpath, '@class', class_selector.class_name)
 
     def xpath_hash(self, id_selector):
         """Translate an ID selector."""
         xpath = self.xpath(id_selector.selector)
-        xpath.add_condition('@id = %s' % xpath_literal(id_selector.id))
+        xpath.add_condition('@id = %s' % self.xpath_literal(id_selector.id))
         return xpath
 
     def xpath_element(self, selector):
@@ -376,15 +378,20 @@ class Translator(object):
         xpath.add_condition('not(%s)' % condition)
         return xpath
 
+    def xpath_contains_function(self, xpath, function):
+        xpath.add_condition('contains(string(.), %s)'
+                            % self.xpath_literal(function.arguments))
+        return xpath
+
+    def function_unsupported(self, xpath, pseudo):
+        raise ExpressionError(
+            "The pseudo-class :%s() is not supported" % pseudo.name)
+
+    xpath_lang_function = function_unsupported
+
 
     # Pseudo: dispatch by pseudo-class name
 
-    def xpath_checked_pseudo(self, xpath):
-        # FIXME: is this really all the elements?
-        xpath.add_condition("(@selected or @checked) and "
-                            "(name(.) = 'input' or name(.) = 'option')")
-        return xpath
-
     def xpath_root_pseudo(self, xpath):
         xpath.add_condition("not(parent::*)")
         return xpath
@@ -434,75 +441,146 @@ class Translator(object):
         xpath.add_condition("not(*) and not(normalize-space())")
         return xpath
 
+    def pseudo_never_matches(self, xpath):
+        """Common implementation for pseudo-classes that never match."""
+        xpath.add_condition("0")
+        return xpath
+
+    xpath_link_pseudo = pseudo_never_matches
+    xpath_visited_pseudo = pseudo_never_matches
+    xpath_hover_pseudo = pseudo_never_matches
+    xpath_active_pseudo = pseudo_never_matches
+    xpath_focus_pseudo = pseudo_never_matches
+    xpath_target_pseudo = pseudo_never_matches
+    xpath_enabled_pseudo = pseudo_never_matches
+    xpath_disabled_pseudo = pseudo_never_matches
+    xpath_checked_pseudo = pseudo_never_matches
 
     # Attrib: dispatch by attribute operator
 
-    def xpath_attrib_exists(self, xpath, name, operator, value):
+    def xpath_attrib_exists(self, xpath, name, value):
         assert not value
         xpath.add_condition(name)
         return xpath
 
-    def xpath_attrib_equals(self, xpath, name, operator, value):
-        xpath.add_condition('%s = %s' % (name, xpath_literal(value)))
+    def xpath_attrib_equals(self, xpath, name, value):
+        xpath.add_condition('%s = %s' % (name, self.xpath_literal(value)))
         return xpath
 
-    def xpath_attrib_different(self, xpath, name, operator, value):
+    def xpath_attrib_different(self, xpath, name, value):
         # FIXME: this seems like a weird hack...
         if value:
             xpath.add_condition('not(%s) or %s != %s'
-                                % (name, name, xpath_literal(value)))
+                                % (name, name, self.xpath_literal(value)))
         else:
             xpath.add_condition('%s != %s'
-                                % (name, xpath_literal(value)))
+                                % (name, self.xpath_literal(value)))
         return xpath
 
-    def xpath_attrib_includes(self, xpath, name, operator, value):
+    def xpath_attrib_includes(self, xpath, name, value):
         xpath.add_condition(
             "%s and contains(concat(' ', normalize-space(%s), ' '), %s)"
-            % (name, name, xpath_literal(' '+value+' ')))
+            % (name, name, self.xpath_literal(' '+value+' ')))
         return xpath
 
-    def xpath_attrib_dashmatch(self, xpath, name, operator, value):
+    def xpath_attrib_dashmatch(self, xpath, name, value):
         # Weird, but true...
         xpath.add_condition('%s and (%s = %s or starts-with(%s, %s))' % (
             name,
-            name, xpath_literal(value),
-            name, xpath_literal(value + '-')))
+            name, self.xpath_literal(value),
+            name, self.xpath_literal(value + '-')))
         return xpath
 
-    def xpath_attrib_prefixmatch(self, xpath, name, operator, value):
+    def xpath_attrib_prefixmatch(self, xpath, name, value):
         xpath.add_condition('%s and starts-with(%s, %s)' % (
-            name, name, xpath_literal(value)))
+            name, name, self.xpath_literal(value)))
         return xpath
 
-    def xpath_attrib_suffixmatch(self, xpath, name, operator, value):
+    def xpath_attrib_suffixmatch(self, xpath, name, value):
         # Oddly there is a starts-with in XPath 1.0, but not ends-with
         xpath.add_condition(
             '%s and substring(%s, string-length(%s)-%s) = %s'
-            % (name, name, name, len(value)-1, xpath_literal(value)))
+            % (name, name, name, len(value)-1, self.xpath_literal(value)))
         return xpath
 
-    def xpath_attrib_substringmatch(self, xpath, name, operator, value):
+    def xpath_attrib_substringmatch(self, xpath, name, value):
         # Attribute selectors are case sensitive
         xpath.add_condition('%s and contains(%s, %s)' % (
-            name, name, xpath_literal(value)))
+            name, name, self.xpath_literal(value)))
         return xpath
 
 
-    # Known but unsupported (functional) pseudo-classes:
+class HTMLTranslator(GenericTranslator):
+    """
+    Translator for HTML documents.
+    """
+    def xpath_checked_pseudo(self, xpath):
+        # FIXME: is this really all the elements?
+        xpath.add_condition(
+            "(@selected and name(.) = 'option') or "
+            "(@checked and name(.) = 'input')")
+        return xpath
 
-    def xpath_unsupported_pseudo(self, xpath, pseudo):
-        raise ExpressionError(
-            "The pseudo-class %r is not supported" % pseudo.name)
+    def xpath_link_pseudo(self, xpath):
+        xpath.add_condition("@href and name(.) = 'a'")
+        return xpath
 
-unsupported = vars(Translator)['xpath_unsupported_pseudo']
-for name in [
-    # Pseudo-elements
-    'first-line', 'first-letter', 'selection', 'before', 'after',
-    # Pseudo-classes
-    'link', 'indeterminate', 'visited', 'active', 'focus', 'hover',
-    # Functional pseudo-classes
-    'target', 'lang', 'enabled', 'disabled',
-    ]:
-    setattr(Translator, name, unsupported)
-del unsupported, name
+    # Links are never visited, the implementation for :visited is the same
+    # as in GenericTranslator
+
+    def xpath_disabled_pseudo(self, xpath):
+        # http://www.w3.org/TR/html5/section-index.html#attributes-1
+        xpath.add_condition('''
+        (
+            @disabled and
+            (
+                name(.) = 'input' or
+                name(.) = 'button' or
+                name(.) = 'select' or
+                name(.) = 'textarea' or
+                name(.) = 'keygen' or
+                name(.) = 'command' or
+                name(.) = 'fieldset' or
+                name(.) = 'optgroup' or
+                name(.) = 'option'
+            )
+        ) or (
+            (
+                name(.) = 'input' or
+                name(.) = 'button' or
+                name(.) = 'select' or
+                name(.) = 'textarea' or
+                name(.) = 'keygen'
+            )
+            and ancestor::fieldset[@disabled]
+        )
+        ''')
+        # FIXME: in the second half, add "and is not a descendant of that
+        # fieldset element's first legend element child, if any."
+        return xpath
+
+    def xpath_enabled_pseudo(self, xpath):
+        # http://www.w3.org/TR/html5/section-index.html#attributes-1
+        xpath.add_condition('''
+        (
+            (
+                name(.) = 'command' or
+                name(.) = 'fieldset' or
+                name(.) = 'optgroup' or
+                name(.) = 'option'
+            )
+            and not(@disabled)
+        ) or (
+            (
+                name(.) = 'input' or
+                name(.) = 'button' or
+                name(.) = 'select' or
+                name(.) = 'textarea' or
+                name(.) = 'keygen'
+            )
+            and not (@disabled or ancestor::fieldset[@disabled])
+        )
+        ''')
+        # FIXME: in the second half, add "and is not a descendant of that
+        # fieldset element's first legend element child, if any."
+        return xpath
